/**
* CMU MSUAV Project (Micro and Small Unmanned Aerial Vehicle)
*          
* \file  main_tracker.cxx
* \brief main function of IMU-aided KLT feature tracker
* \date  18-Apr-2008
*
* \author Myung Hwangbo (myung@cs.cmu.edu) and Jun-Sik Kim (kimjs@cs.cmu.edu)
*
* Copyright (c) 2006-2009 Myung Hwangbo and Jun-Sik Kim
* Robotics Institute, Carnegie Mellon University
*
*/

#ifdef _WIN32
#include <windows.h>
#include <conio.h>
#endif

#include <iostream>
#include <fstream>
#include <sstream>
#include <cstdlib>

#include <string>
#include <ipp.h>

#include <execinfo.h>
#include <signal.h>
#include <stdlib.h>
#include <ucontext.h>
#include <stdexcept>
#include <segvcatch.h>
#include <malloc.h>

#include "ctimer.h"
#include "featurePool.h"
#include "playback_tracker.h"
#include "display_tracker.h"

#include <cuda_runtime.h>

#include "litmus.h"

//#include <blitz/array.h>

#include <boost/interprocess/managed_shared_memory.hpp>
#include <boost/interprocess/sync/interprocess_semaphore.hpp>
#include <boost/interprocess/sync/interprocess_barrier.hpp>

using namespace std;
using namespace boost::interprocess;

#define RESET_RELEASE_ON_MISS

CFeature2DPool* featurePool = NULL;

int min_feature = 400;
int max_feature = 400;
int templ_size = 11;
int mask_size = 10;
int pyramid_level = 3;

double disp_scale = 1.0;
int img_width = 640;
int img_height = 480;
int t_offset = 0;
char track_log_name[256];
char imu_shape_name[256];
char imu_list_name[256];
char img_list_name[256];
char camera_param_name[256];
char camera_table_name[256];

enum eMode
{
//	LINUX = 0,
//	RT_LINUX = 1,
	LITMUS_RT = 2
};

eMode gMode = LITMUS_RT;

int sched_priority = sched_get_priority_min(SCHED_FIFO);


bool b_display = false;
bool b_gpu = false;
bool b_imu = true;
bool b_stop = false;
bool b_paused = false;
bool b_log = false;
bool b_run = true;

bool b_preload = true;


void exit_handler();
bool read_input_arg(int argc, char* argv[]);
bool read_config_file(const char* fname);

//


// default: 10ms wcet w/ 30 fps display
float wcet_ms = 10;
float period_ms = 33;

uint64_t period_ns = period_ms * __NS_PER_MS;

int numGPUUsers = -1;

int numTotalTasks = -1;

int numGPUPartitions = 1;
int gpuPartition = 0;
int numGPUs = 1;

bool b_waitForRelease = false;
int initCPU = -1;
int numFramesToProcess = -1;


extern "C" bool ExitCUDA(void);


inline uint64_t timespec_to_ns(const struct timespec& t)
{
	return(t.tv_sec*1e9 + t.tv_nsec);
}

inline struct timespec ns_to_timespec(const uint64_t& ns)
{
	struct timespec temp = {ns/1e9, ns - ns/1e9};
	return(temp);	
}

inline uint64_t clock_gettime_ns(clockid_t clk_id)
{
	struct timespec temp;
	clock_gettime(clk_id, &temp);
	return timespec_to_ns(temp);
}





enum eCheckPoint
{
	INIT = 0,
	LOAD = 1,
	JOB = 2,
	EXIT_1 = 3,
	EXIT_2 = 4,
	EXIT_3 = 5,
	EXIT_4 = 6,
	EXIT_5 = 7,
	EXIT_6 = 8
};

//int //chpt = INIT;

//int sigsegved = 0;

//typedef struct ucontext sig_ucontext_t;
//
//void crit_err_hdlr(int sig_num, siginfo_t *info, void *ucontext)
//{
//	void *             array[50];
//	void *             caller_address;
//	char **            messages;
//	int                size, i;
//	sig_ucontext_t *   uc;
//	
//	uc = (sig_ucontext_t *)ucontext;
//	
//	/* Get the address at the time the signal was raised from the EIP (x86) */
//	caller_address = (void *) uc->uc_mcontext.gregs[REG_RIP];   
//
//	fprintf(stderr, "%d: signal %d (%s), address is %p from %p (check point = %d)\n", 
//			getpid(), sig_num, strsignal(sig_num), info->si_addr, 
//			(void *)caller_address, //chpt);
//	
//	size = backtrace(array, 50);
//	
//	/* overwrite sigaction with caller's address */
//	array[1] = caller_address;
//	
//	messages = backtrace_symbols(array, size);
//	
//	/* skip first stack frame (points here) */
//	for (i = 1; i < size && messages != NULL; ++i)
//	{
//		fprintf(stderr, "%d: [bt]: (%d) %s\n", getpid(), i, messages[i]);
//	}
//	
//	free(messages);
//
//	//char buff[256] = {0};
//	//sprintf(buff, "xterm -e gdb attach %d", getpid());
//	//system(buff);
//	
//	//exit(EXIT_FAILURE);
//
//
//	++(uc->uc_mcontext.gregs[REG_RIP]);  // DANGER, DANGER, DANGER!
//
//	sigsegved = 1;
//}
//
//
//
//
//void handler(int sig)
//{
//	void *array[10];
//	size_t size;
//	size = backtrace(array, 10);
//	fprintf(stderr, "Error: signal %d:\n", sig);
//	backtrace_symbols_fd(array, size, 2);
//	exit(1);
//}

////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////

#if 0
void init_linux()
{
    mlockall(MCL_CURRENT | MCL_FUTURE);
	
    // set the affinity
    cpu_set_t *affinity = CPU_ALLOC(12);
    size_t affinitySize = CPU_ALLOC_SIZE(12);
    CPU_ZERO_S(affinitySize, affinity);
    for(int i = initCPU; i < initCPU+6; ++i)
    {   
        CPU_SET_S(i, affinitySize, affinity);
    }   
    assert(0 == sched_setaffinity(getpid(), affinitySize, affinity));
    CPU_FREE(affinity);
	
	if (gMode == RT_LINUX)
	{
		// let's go FIFO
		struct sched_param fifoParams;
		memset(&fifoParams, 0, sizeof(fifoParams));
		fifoParams.sched_priority = sched_priority;
		assert(0 == sched_setscheduler(getpid(), SCHED_FIFO, &fifoParams));
	}
}


int run_with_linux()
{
	init_linux();
	
	period_ns = period_ms * __NS_PER_MS;
	
	//CTimer timer;
	//double t0 = 0, t1 = 0, t2 = 0, t3 = 0, t4 = 0, t5 = 0;
	double KK[9], Ki[9];
	
	CvSize img_size = cvSize(img_width, img_height);
	CCameraModel* cam = new CCameraModel(camera_param_name, camera_table_name);
	
	featurePool = CFeature2DPool::CreateCFeature2DPool(img_size, templ_size, pyramid_level, mask_size, b_gpu, cam,
													   numGPUs, gpuPartition, numGPUPartitions, numGPUUsers, LINUX_SEM);
	
	if (b_display) init_track_display(disp_scale);
	
	std::vector<IplImage*> preloadVid(0);
	if(b_preload) {
		preloadVid.reserve(numFramesToProcess);
		
		// the feature pool destructor will clean up the loaded images
		// leaving the vector full of dangling pointers.  >:P
		int numFrames = preload(img_list_name, img_size, preloadVid, numFramesToProcess);
		if(numFrames > 0) {
			printf("Preloaded %d frames.\n", numFrames);
		}
		else {
			if(numFrames < 0) {
				printf("Failed to preload video stream!\n");
			}
			else if(numFrames == 0) {
				printf("No frames were loaded!\n");
			}
			exit(-1);
		}
	}
	
	uint64_t *init_release_time = NULL;
	managed_shared_memory *segment_ptr = NULL;
	if(numTotalTasks != -1)
	{
		printf("%d creating shared memory for release time\n", getpid());
		shared_memory_object::remove("init_release_memory");
		segment_ptr = new managed_shared_memory(create_only, "init_release_memory", 4*1024);
		
		init_release_time = segment_ptr->construct<uint64_t>("uint64_t instance")();
		*init_release_time = 0;
	}
	else
	{
		do
		{
			try
			{
				printf("%d opening init_release_memory shared memory\n", getpid());
				segment_ptr = new managed_shared_memory(open_only, "init_release_memory");
			}
			catch(...)
			{
				printf("%d shared init_release_memory memory not ready.  sleeping\n", getpid());
				sleep(1);
			}
		}while(segment_ptr == NULL);
		
		init_release_time = segment_ptr->find<uint64_t>("uint64_t instance").first;
	}		
	
	
	// linux sync release.
	barrier *release_barrier = NULL;	
	struct timespec *sync_release_ptr = NULL;
	if(numTotalTasks != -1)
	{
		printf("%d creating release barrier for %d users\n", getpid(), numTotalTasks);
		release_barrier = segment_ptr->construct<barrier>("barrier release_barrier")(numTotalTasks);
		
		printf("%d creating sync timespec\n", getpid());
		sync_release_ptr = segment_ptr->construct<struct timespec>("struct timespec sync_release")();
		
		//printf("--> %s:%d\n", __FUNCTION__, __LINE__);
	}
	else
	{
		release_barrier = segment_ptr->find<barrier>("barrier release_barrier").first;
		sync_release_ptr = segment_ptr->find<struct timespec>("struct timespec sync_release").first;
	}
	
	
	struct timespec spec;
	uint64_t release;
	uint64_t finish;
//	blitz::Array<uint64_t, 1> responseTimeLog(numFramesToProcess);
	
	//printf("--> %s:%d\n", __FUNCTION__, __LINE__);
	
	//
	// Main loop
	//
	int ctime = 0, ptime = 0, frame = 0, frame_idx = 0;
	
	if (b_waitForRelease) {
		
		//printf("--> %s:%d\n", __FUNCTION__, __LINE__);
		
		sleep(5);
		
		//printf("--> %s:%d\n", __FUNCTION__, __LINE__);
		
		release_barrier->wait();
		
		//printf("--> %s:%d\n", __FUNCTION__, __LINE__);
		
		if(numTotalTasks != -1)
        {   
            printf("%d: setting up sync release\n", getpid());
            clock_gettime(CLOCK_MONOTONIC, &spec);
            spec.tv_sec += 30; 
            *sync_release_ptr = spec;
        }
		
		//printf("--> %s:%d\n", __FUNCTION__, __LINE__);
		
        release_barrier->wait();
		
		//printf("--> %s:%d\n", __FUNCTION__, __LINE__);
		
        spec = *sync_release_ptr;
		
		release = timespec_to_ns(spec);
		
		printf("%d: sleeping until %lu for sync release.\n", getpid(), release);
		clock_nanosleep(CLOCK_MONOTONIC, TIMER_ABSTIME, &spec, NULL);
		
		//printf("--> %s:%d\n", __FUNCTION__, __LINE__);
	}
	else
	{
		release = clock_gettime_ns(CLOCK_MONOTONIC);
	}
	
	//printf("--> %s:%d\n", __FUNCTION__, __LINE__);

	while(b_run)
	{
		//timer.Start();
	
		//chpt = 1000;

		if (b_display) {
			char ch = (char)cvWaitKey(5);
			if (ch == 'q') break;
			else if (ch == 'c') int ret = system("cls");
			else if (ch > 0) b_paused = !b_paused;
		}
		
		//chpt = 1010;

		if (b_paused) {
			continue; //Sleep(10);
		}
		
		//printf("--> %s:%d\n", __FUNCTION__, __LINE__);
		
		//
		// Step 1: Read the data
		//
		IplImage* gimgf = NULL;
		if(!b_preload) {
			gimgf = cvCreateImage(img_size, IPL_DEPTH_32F, 1);
			ctime = playback(img_list_name, t_offset, gimgf);
			if (ctime < 0) break;
		}
		else {
			
			//printf("--> %s:%d\n", __FUNCTION__, __LINE__);
			
			if(numFramesToProcess != -1) {
				if(frame >= numFramesToProcess)
					break;
				
				if(frame_idx >= preloadVid.size())
					frame_idx = 0;
			}
			else {
				if(frame >= preloadVid.size())
					break;
			}
			
			
			//printf("--> %s:%d\n", __FUNCTION__, __LINE__);
			
			gimgf = preloadVid[frame_idx];
			ctime = 0;
		}
		
		//chpt = 1020;

		//printf("--> %s:%d\n", __FUNCTION__, __LINE__);		
		
		featurePool->AddImage(frame, gimgf);

		//chpt = 1030;
		//t0 = timer.Stop();
		
		//printf("--> %s:%d\n", __FUNCTION__, __LINE__);
		
		//
		// Step 2: Predict the camera rotation from the IMU
		//
		double H[9] = {1, 0, 0, 0, 1, 0, 0, 0, 1};
		if (b_imu && ptime != 0) {
			if (!compute_homography_IMU(ptime, ctime, imu_list_name, imu_shape_name, 
										cam->GetKK(KK), cam->GetKi(Ki), H)) break;
		}

		//chpt = 1040;
		
		//
		// Step 3: Track and select the features
		//
		if (b_gpu) {
			//printf("--> %s:%d\n", __FUNCTION__, __LINE__);
			featurePool->Track_GPU(frame, gimgf, H);							//t1 = timer.Stop();
			
			//printf("--> %s:%d\n", __FUNCTION__, __LINE__);			
			
			featurePool->Select_GPU(frame, min_feature, max_feature, gimgf, false);	//t2 = timer.Stop();
		}
		else {
			//chpt = 1050;
			featurePool->Track(frame, ctime, gimgf, (b_imu == true) ? H : NULL); //t1 = timer.Stop();
			//chpt = 1060;
			featurePool->Select(frame, min_feature, max_feature, gimgf, false);	 //t2 = timer.Stop();
			//chpt = 1070;
		}
		
		//printf("--> %s:%d\n", __FUNCTION__, __LINE__);		
		//chpt = 1080;
		featurePool->Purge(frame, 10);											//t3 = timer.Stop();
		//chpt = 1090;
		if (b_log) featurePool->SaveToFile(0, track_log_name);
		
		//
		// Step 4: Display
		//
		if (b_display) run_track_display();									//t4 = timer.Stop();
		
		//chpt = 1100;

		/*
		 CFeatureTrackStat& stat = featurePool->GetStat();
		 printf("%3d, %8d %3d | %3d = %3d/%3d %3d | %5.1f = %5.1f + %5.1f + %5.1f, %5.1f\n",
		 stat.frame, ctime-t_offset, (ctime-ptime), stat.n_total, stat.n_success, stat.n_total - stat.n_success, stat.n_new,
		 (t3 - t0)*1e3, (t1 - t0)*1e3, (t2 - t1)*1e3, (t3 - t2)*1e3, (t0 - t5)*1e3);
		 */
		
		//
		// Step 3: Update
		//
		if (b_stop) b_paused = true;
		
		ptime = ctime;															//t5 = timer.Stop();
		
//		clock_gettime(CLOCK_MONOTONIC, &spec);
//		finish = timespec_to_ns(spec);
		
		//chpt = 1110;

		//printf("period = %lu\n", period_ns);		
		//printf("release = %lu\n", release);
		//printf("finish = %lu\n", finish);
		//printf("response time = %lu\n", finish - release);
//		responseTimeLog(frame_idx) = finish - release;		
		
		/*
		 printf("Image = %f, Track = %f, Select = %f, Purge = %f, Display = %f, Total = %f\n",
		 1e3*t0, 1e3*(t1 - t0), 1e3*(t2 - t1), 1e3*(t3 - t2), 1e3*(t4 - t3), 1e3*t5);
		 */
		
		
		frame++;
		frame_idx++;
		
		if (release + period_ns < finish) {
			// miss
#ifdef RESET_RELEASE_ON_MISS
			release = finish;
#else
			release += period_ns;
#endif
			sched_yield();  // release already passed.  yield to other jobs if need be.
		}
		else {
			release += period_ns;
			spec = ns_to_timespec(release);
			clock_nanosleep(CLOCK_MONOTONIC, TIMER_ABSTIME, &spec, NULL);
		}

//		// TODO: Following still needed in Linux?
//		// litmus has more accurate timing than we do, so if we're running earlier than
//		// estimated, adjust our release time to be as close as possible to the actual
//		// release time.
//		clock_gettime(CLOCK_MONOTONIC, &spec);
//		release = min(timespec_to_ns(spec), release);

		//chpt = 1120;
	}
	
	/*	   
	 extern double RUNTIMES[500];
	 extern int RTCOUNT;
	 
	 double sum = 0.0;
	 for(int i = 0; i < RTCOUNT; ++i)
	 {
	 sum += RUNTIMES[i];
	 }
	 //fprintf(stderr, "--> %p\n", &RUNTIMES[0]);
	 fprintf(stderr, "Track kernels avg = %f\n", sum/RTCOUNT);
	 */
	
//	char gpu_using_str[] = "*\n";
//	char cpu_only_str[] = "\n";
//#define USED(arr) ((arr)(blitz::Range(blitz::fromStart,frame_idx-1)))
//	printf("DONE,%d,%d,%f,%f,%f,%lu,%lu,%f,%d,%d,%d,==,%s",
//		   getpid(),
//		   initCPU,
//		   period_ms,
//		   // average
//		   blitz::mean(USED(responseTimeLog)),
//		   // average pct of period
//		   100.0*(blitz::mean(USED(responseTimeLog))/period_ns),    
//		   // min
//		   blitz::min(USED(responseTimeLog)),
//		   // max
//		   blitz::max(USED(responseTimeLog)),
//		   // average slack
//		   blitz::mean(period_ns - USED(responseTimeLog)),
//		   // num jobs
//		   frame_idx,
//		   // num misses
//		   blitz::count(USED(responseTimeLog) > period_ns),
//		   // num misses w/ unbounded
//		   blitz::count(USED(responseTimeLog) > 2*period_ns),
//		   // flag gpu-using tasks
//		   ((b_gpu) ? gpu_using_str : cpu_only_str));

	if(b_gpu)
	{
		printf("exiting cuda...\n");
	 	ExitCUDA();
	}
    
	printf("cleaning up feature pool.\n");
    //CFeature2DPool::DestroyCFeature2DPool(&featurePool);
	featurePool = NULL;
	  
	if (b_display) close_track_display();
	if (cam) delete cam;
	
	return 0;
}
#endif

////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////


int run_with_litmus()
{
	period_ns = period_ms * __NS_PER_MS;

	if(sporadic_task_ns((lt_t)(wcet_ms * __NS_PER_MS), (lt_t)period_ns, 0, initCPU, RT_CLASS_SOFT, NO_ENFORCEMENT, 1) < 0)
	{
		printf("sporadic_task_ns() failed.\n");
		exit(-1);
	}
	init_litmus();
	
	// need to change mode before we init the GPUs.
	if (task_mode(LITMUS_RT_TASK) != 0)
	{
		printf("could not switch to RT task.\n");
		exit(-1);
	}		
	
	
	//CTimer timer;
	//double t0 = 0, t1 = 0, t2 = 0, t3 = 0, t4 = 0, t5 = 0;
	double KK[9], Ki[9];
	
	CvSize img_size = cvSize(img_width, img_height);
	CCameraModel* cam = new CCameraModel(camera_param_name, camera_table_name);

	featurePool = CFeature2DPool::CreateCFeature2DPool(
		img_size, templ_size, pyramid_level, mask_size, b_gpu, cam,
		numGPUs, gpuPartition, numGPUPartitions, numGPUUsers, LITMUS);
	
	if (b_display) init_track_display(disp_scale);

	
	//chpt = LOAD;
	std::vector<IplImage*> preloadVid(0);
	if(b_preload) {
		preloadVid.reserve(numFramesToProcess);
		
		// the feature pool destructor will clean up the loaded images
		// leaving the vector full of dangling pointers.  >:P
		int numFrames = preload(img_list_name, img_size, preloadVid, numFramesToProcess);
		if(numFrames > 0) {
			printf("Preloaded %d frames.\n", numFrames);
		}
		else {
			if(numFrames < 0) {
				printf("Failed to preload video stream!\n");
			}
			else if(numFrames == 0) {
				printf("No frames were loaded!\n");
			}
			exit(-1);
		}
	}
	
	uint64_t *init_release_time = NULL;
	managed_shared_memory *segment_ptr = NULL;
	if(numTotalTasks != -1)
	{
		printf("%d creating shared memory for release time\n", getpid());
		shared_memory_object::remove("init_release_memory");
		segment_ptr = new managed_shared_memory(create_only, "init_release_memory", 4*1024);
		
		init_release_time = segment_ptr->construct<uint64_t>("uint64_t instance")();
		*init_release_time = 0;
	}
	else
	{
		do
		{
			try
			{
				printf("%d opening init_release_memory shared memory\n", getpid());
				segment_ptr = new managed_shared_memory(open_only, "init_release_memory");
			}
			catch(...)
			{
				printf("%d shared init_release_memory memory not ready.  sleeping\n", getpid());
				sleep(1);
			}
		}while(segment_ptr == NULL);
		
		init_release_time = segment_ptr->find<uint64_t>("uint64_t instance").first;
	}		
	
//	struct timespec spec;
//	uint64_t release;
//	uint64_t finish;
//	blitz::Array<uint64_t, 1> responseTimeLog(numFramesToProcess);
	
	
	//chpt = JOB;
	
	//
	// Main loop
	//
	int ctime = 0, ptime = 0, frame = 0, frame_idx = 0;
	
	if (b_waitForRelease) {
		int ret = wait_for_ts_release();
		if (ret != 0)
		{
			printf("wait_for_ts_release() failed");
			exit(-1);
		}
	}
	else
	{
		sleep_next_period();
	}

//	clock_gettime(CLOCK_MONOTONIC, &spec);
	
//	// record the time of the earliest guy to start
//	release = timespec_to_ns(spec);
//	if (!__sync_bool_compare_and_swap(init_release_time, 0, release))
//	{
//		release = *init_release_time;
//	}
	
	while(b_run)
	{
//		printf("STARTING FRAME\n");
		//timer.Start();

		//chpt = 1000;
		
		if (b_display) {
			char ch = (char)cvWaitKey(5);
			if (ch == 'q') break;
			else if (ch == 'c') int ret = system("cls");
			else if (ch > 0) b_paused = !b_paused;
		}
		
		if (b_paused) {
			continue; //Sleep(10);
		}

		//chpt = 1010;
		
		//
		// Step 1: Read the data
		//
		IplImage* gimgf = NULL;
		if(!b_preload) {
			gimgf = cvCreateImage(img_size, IPL_DEPTH_32F, 1);
			ctime = playback(img_list_name, t_offset, gimgf);
			if (ctime < 0) break;
		}
		else {
			if(numFramesToProcess != -1) {
				if(frame >= numFramesToProcess) {
					b_run = false;
					break;
				}
				
				if(frame_idx >= preloadVid.size()) {
					frame_idx = 0;
				}
			}
			else {
				if(frame >= preloadVid.size()) {
					b_run = false;
					break;
				}
			}
			
			gimgf = preloadVid[frame_idx];
			ctime = 0;
		}
		
		//chpt = 1020;

		featurePool->AddImage(frame, gimgf);

		//chpt = 1030;
		//t0 = timer.Stop();
		
		//
		// Step 2: Predict the camera rotation from the IMU
		//
		double H[9] = {1, 0, 0, 0, 1, 0, 0, 0, 1};
		if (b_imu && ptime != 0) {
			if (!compute_homography_IMU(ptime, ctime, imu_list_name, imu_shape_name, 
										cam->GetKK(KK), cam->GetKi(Ki), H)) break;
		}

		//chpt = 1040;
		
		//
		// Step 3: Track and select the features
		//
		//chpt = 1050;
		featurePool->Track_GPU(frame, gimgf, H);							//t1 = timer.Stop();
		//chpt = 1060;
		featurePool->Select_GPU(frame, min_feature, max_feature, gimgf, false);	//t2 = timer.Stop();
		//chpt = 1070;

		
		//chpt = 1080;
		featurePool->Purge(frame, 10);											//t3 = timer.Stop();
		//chpt = 1090;
		if (b_log) featurePool->SaveToFile(0, track_log_name);
		
		//
		// Step 4: Display
		//
		if (b_display) run_track_display();									//t4 = timer.Stop();
		
		/*
		 CFeatureTrackStat& stat = featurePool->GetStat();
		 printf("%3d, %8d %3d | %3d = %3d/%3d %3d | %5.1f = %5.1f + %5.1f + %5.1f, %5.1f\n",
		 stat.frame, ctime-t_offset, (ctime-ptime), stat.n_total, stat.n_success, stat.n_total - stat.n_success, stat.n_new,
		 (t3 - t0)*1e3, (t1 - t0)*1e3, (t2 - t1)*1e3, (t3 - t2)*1e3, (t0 - t5)*1e3);
		 */
		
		//
		// Step 3: Update
		//
		if (b_stop) b_paused = true;
		
		ptime = ctime;															//t5 = timer.Stop();
		
//		clock_gettime(CLOCK_MONOTONIC, &spec);
//		finish = timespec_to_ns(spec);
		
		//printf("period = %lu\n", period_ns);		
		//printf("release = %lu\n", release);
		//printf("finish = %lu\n", finish);
		//printf("response time = %lu\n", finish - release);
//		responseTimeLog(frame_idx) = finish - release;		
		
		/*
		 printf("Image = %f, Track = %f, Select = %f, Purge = %f, Display = %f, Total = %f\n",
		 1e3*t0, 1e3*(t1 - t0), 1e3*(t2 - t1), 1e3*(t3 - t2), 1e3*(t4 - t3), 1e3*t5);
		 */
		
		// this is an estimated upper-bound on release time.  it may be off by several microseconds.
//#ifdef RESET_RELEASE_ON_MISS
//		release = (release + period_ns < finish) ?
//			finish :  /* missed deadline.  adopt next release as current time. */
//			release + period_ns;  /* some time in the future. */
//#else
//		release = release + period_ns; // allow things to get progressively later.
//#endif

		//printf("next release = %lu\n\n", release);

		
		
		frame++;
		frame_idx++;
		
		if(frame%200 == 0)
			printf("[%d] completed frame %d\n", getpid(), frame);
		
		sleep_next_period();
		
		// litmus has more accurate timing than we do, so if we're running earlier than
		// estimated, adjust our release time to be as close as possible to the actual
		// release time.
//		clock_gettime(CLOCK_MONOTONIC, &spec);
//		release = min(timespec_to_ns(spec), release);

		//chpt = 1100;
		
//		printf("ENDING FRAME\n");
	}
	
	if(task_mode(BACKGROUND_TASK) != 0)
		printf("could not switch to background task??\n");

	//chpt = EXIT_1;
	
	/*	   
	 extern double RUNTIMES[500];
	 extern int RTCOUNT;
	 
	 double sum = 0.0;
	 for(int i = 0; i < RTCOUNT; ++i)
	 {
	 sum += RUNTIMES[i];
	 }
	 //fprintf(stderr, "--> %p\n", &RUNTIMES[0]);
	 fprintf(stderr, "Track kernels avg = %f\n", sum/RTCOUNT);*/

//	char gpu_using_str[] = "*\n";
//	char cpu_only_str[] = "\n";
//#define USED(arr) ((arr)(blitz::Range(blitz::fromStart,frame_idx-1)))
//	printf("DONE,%d,%d,%f,%f,%f,%lu,%lu,%f,%d,%d,%d,==,%s",
//		   getpid(),
//		   initCPU,
//		   period_ms,
//		   // average
//		   blitz::mean(USED(responseTimeLog)),
//		   // average pct of period
//		   100.0*(blitz::mean(USED(responseTimeLog))/period_ns),    
//		   // min
//		   blitz::min(USED(responseTimeLog)),
//		   // max
//		   blitz::max(USED(responseTimeLog)),
//		   // average slack
//		   blitz::mean(period_ns - USED(responseTimeLog)),
//		   // num jobs
//		   frame_idx,
//		   // num misses
//		   blitz::count(USED(responseTimeLog) > period_ns),
//		   // num misses w/ unbounded
//		   blitz::count(USED(responseTimeLog) > 2*period_ns),
//		   // flag gpu-using tasks
//		   ((b_gpu) ? gpu_using_str : cpu_only_str));
//
//    //chpt = EXIT_2;

	if(b_gpu)
	{
		printf("exiting cuda...\n");
		ExitCUDA();
	}

    //chpt = EXIT_3;

	printf("cleaning up feature pool.\n");
	//CFeature2DPool::DestroyCFeature2DPool(&featurePool);
	featurePool = NULL;

    //chpt = EXIT_4;

	if (b_display) close_track_display();
	if (cam) delete cam;
	
    //chpt = EXIT_5;

	return 0;
}


////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////

class ippException: public exception
{
public:
	const char* what() const throw()
	{
		extern Ipp32f *ex_src;
		extern Ipp32f *ex_dst;
		extern int ex_srcStep;
		extern int ex_dstStep;
		extern IppiSize ex_size;
		extern IppiRect ex_src_roi;
		extern IppiRect ex_dst_roi;
		extern double ex_coeffs[3][3];
		extern int ex_mofolevel;
		extern int ex_bad_ptr;

		sprintf((char*)&(blarg[0]), "src: %p, (%d, %d), %d, (%d, %d, %d, %d),\n"
					   "dst (level = %d): %p, %d, (%d, %d, %d, %d)\n"
					   "guess = %d\n",
					   ex_src, ex_size.width, ex_size.height, ex_srcStep, ex_src_roi.x, ex_src_roi.y, ex_src_roi.width, ex_src_roi.height,
					   ex_mofolevel, ex_dst, ex_dstStep, ex_dst_roi.x, ex_dst_roi.y, ex_dst_roi.width, ex_dst_roi.height,
					   ex_bad_ptr);

		//ippiWarpPerspectiveBack_32f_C1R(ex_src, ex_size, ex_srcStep, ex_src_roi, ex_dst, ex_dstStep, ex_dst_roi, ex_coeffs, IPPI_INTER_LINEAR);

		return((const char*)&blarg[0]);

	}
private:
	char blarg[256];
};

void myhandler()
{
	printf("\n------------\n");
	throw ippException();
	printf("\n------------\n");
}

int main(int argc, char* argv[])
{
//	struct sigaction sigact;
//	
//	sigact.sa_sigaction = crit_err_hdlr;
//	sigact.sa_flags = SA_RESTART | SA_SIGINFO;	

	//mallopt(M_MMAP_THRESHOLD, 0);
//	sigaction(SIGSEGV, &sigact, (struct sigaction *)NULL);

	//	segvcatch::init_segv(myhandler);

	//signal(SIGSEGV, handler);
	atexit(exit_handler);
	if (!read_input_arg(argc, argv)) return 0;

	int ret = 0;
	switch(gMode)
	{
//		case LINUX:
//			ret = run_with_linux();
//			break;
//		case RT_LINUX:
//			ret = run_with_linux();
//			break;
		case LITMUS_RT:
			ret = run_with_litmus();
			break;
		default:
			"UNSUPPORTED MODE!\n";
			exit(-1);
			break;
	}

	//printf("all done!\n");

    //chpt = EXIT_6;

//	if(sigsegved)
//		ret = EXIT_FAILURE;

	return(ret);
}

void exit_handler()
{
	//printf("exit() is called. \n");
    b_run = false;

	if (featurePool)
	{
		featurePool->cleanUpLocks();
	}
}





extern bool ENABLE_AFFINITY;
extern bool RELAX_FIFO_MAX_LEN;
extern bool USE_DYNAMIC_GROUP_LOCKS;
extern bool ENABLE_CHUNKING;
extern bool MIGRATE_VIA_SYSMEM;
extern size_t CHUNK_SIZE;

extern int GPU_PARTITION;
extern int GPU_PARTITION_SIZE;
extern int NUM_SIMULT_USERS;
extern int UNCONSTRAINED_ACCESS;

			
bool read_input_arg(int argc, char* argv[])
{	
	for (int i=1; i < argc; i++) 
	{
		if      (strncmp(argv[i], "-log",   4) == 0)  b_log = true;
		else if (strncmp(argv[i], "-gpu",   4) == 0)  b_gpu = true;
		else if (strncmp(argv[i], "-cpu",   4) == 0)  b_gpu = false;
		else if (strncmp(argv[i], "-stop",  5) == 0)  b_stop = true;
		else if (strncmp(argv[i], "-f",     2) == 0) {
			if ((i+1) >= argc) return false;
			if (!read_config_file(argv[i+1])) return false; i++;
		}
		else if (strncmp(argv[i], "-disp", 5) == 0) {
			if ((i+1) >= argc) return false;
			disp_scale = atof(argv[i+1]); i++;
			b_display = true;
		}
		else if (strncmp(argv[i], "-imu", 4) == 0) {
			if ((i+1) >= argc) return false;
			t_offset = atoi(argv[i+1]); i++;
			b_imu = true;
		}
		else if (strncmp(argv[i], "-n", 2) == 0) {
			if ((i+1) >= argc) return false;
			max_feature = min_feature = atof(argv[i+1]); i++;
		}
		else if (strncmp(argv[i], "-t", 2) == 0) {
			if ((i+1) >= argc) return false;
			templ_size = 2*atoi(argv[i+1]) + 1; i++;
		}
//		else if (strncmp(argv[i], "-linux", 6) == 0) {
//			gMode = LINUX;
//		}
//		else if (strncmp(argv[i], "-rtlinux", 8) == 0) {
//			gMode = RT_LINUX;
//		}
		else if (strncmp(argv[i], "-P", 2) == 0) {
			if ((i+1) >= argc) return false;
			sched_priority = atoi(argv[i+1]); i++;			
		}
		else if (strncmp(argv[i], "-T", 2) == 0) {
			if ((i+1) >= argc) return false;
			numTotalTasks = atoi(argv[i+1]); i++;			
		}		

		else if (strncmp(argv[i], "-litmus", 7) == 0) {
			gMode = LITMUS_RT;
		}
		else if (strncmp(argv[i], "-N", 2) == 0) {
			if ((i+1) >= argc) return false;
			numGPUUsers = atoi(argv[i+1]); i++;			
		}
		else if (strncmp(argv[i], "-w", 2) == 0) {
			b_waitForRelease = true;
		}
		else if (strncmp(argv[i], "-GZ", 3) == 0) {
			if ((i+1) >= argc) return false;
			numGPUPartitions = atoi(argv[i+1]); i++;						
		}
		else if (strncmp(argv[i], "-GP", 3) == 0) {
			if ((i+1) >= argc) return false;
			gpuPartition = atoi(argv[i+1]); i++;
			GPU_PARTITION = gpuPartition;
		}
		else if (strncmp(argv[i], "-g", 2) == 0) {
			if ((i+1) >= argc) return false;
			numGPUs = atoi(argv[i+1]); i++;
			GPU_PARTITION_SIZE = numGPUs;
		}				
		else if (strncmp(argv[i], "-p", 2) == 0) {
			if ((i+1) >= argc) return false;
			initCPU = atoi(argv[i+1]); i++;						
		}
		else if (strncmp(argv[i], "-R", 2) == 0) {
			if ((i+1) >= argc) return false;
			period_ms = 1000.0/atoi(argv[i+1]); i++;						
		}		
		else if (strncmp(argv[i], "-F", 2) == 0) {
			if ((i+1) >= argc) return false;
			numFramesToProcess = atoi(argv[i+1]); i++;						
		}
		else if (strncmp(argv[i], "-affinity", sizeof("-affinity")-1) == 0) {
			ENABLE_AFFINITY = true;
		}
		else if (strncmp(argv[i], "-relax_fifo", sizeof("-relax_fifo")-1) == 0) {
			RELAX_FIFO_MAX_LEN = true;
		}
		else if (strncmp(argv[i], "-dgl", sizeof("-dgl")-1) == 0) {
			USE_DYNAMIC_GROUP_LOCKS = true;
		}
		else if (strncmp(argv[i], "-sysmem", sizeof("sysmem")-1) == 0) {
			MIGRATE_VIA_SYSMEM = true;
		}
		else if (strncmp(argv[i], "-chunk", sizeof("-chunk")-1) == 0) {
			ENABLE_CHUNKING = true;
			
			if ((i+1) >= argc) return false;
			CHUNK_SIZE = atoi(argv[i+1])*1024;  // interpret input as K. convert to bytes.
			i++;
		}
		else if (strncmp(argv[i], "-engine_locks", sizeof("-engine_locks")-1) == 0) {
			NUM_SIMULT_USERS = 2;  // one ee lock, one ce lock.
		}
		else if (strncmp(argv[i], "-u", 2) == 0) {
			UNCONSTRAINED_ACCESS = true;
		}

		else {
			printf("Invalid switch: %s\n\n", argv[i]);
			printf("usage: %s [-f config] [-log] [-gpu] [-stop] [-imu time_ms] [-n num] [-t size] [-disp scale]\n\n", argv[0]);
			return false;
		}
	}

#if 1
	printf("\n[Setting]\n");
	printf(" GPU          : %s\n", b_gpu      ? "ON" : "OFF");
	printf(" IMU          : %s\n", b_imu      ? "ON" : "OFF");
	printf(" Log          : %s\n", b_log      ? "ON" : "OFF");
	printf(" Stop         : %s\n", b_stop     ? "ON" : "OFF");
	printf(" Display      : %s\n", b_display  ? "ON" : "OFF");
	printf(" Time_offset  : %d\n", t_offset);
	printf(" Disp_scale   : %.1f\n", disp_scale);
	printf(" Templ_size   : %d\n", templ_size);
	printf(" Mask_size    : %d\n", templ_size);
	printf(" Feature_num  : %d\n", max_feature);
	printf(" Image_width  : %d\n", img_width);
	printf(" Image_heigh  : %d\n", img_height);
	printf(" Image_data   : %s\n", img_list_name);
	printf(" IMU_data     : %s\n", imu_list_name);
	printf(" IMU_shape    : %s\n", imu_shape_name);
	printf(" Camera_param : %s\n", camera_param_name);
	printf(" Camera_table : %s\n", camera_table_name);
	printf(" Log_filename : %s\n", track_log_name);

	
	
	printf(" WCET(ms)     : %f\n", wcet_ms);
	printf(" Period(ms)   : %f\n", period_ms);
	printf(" Num GPUs     : %d\n", numGPUs);
	printf(" GPU Partition: %d\n", gpuPartition);
	printf(" GPU Partition#: %d\n", numGPUPartitions);
	printf(" Num GPU Users: %d\n", numGPUUsers);
	printf(" Wait Release : %d\n", (int)b_waitForRelease);

	printf(" Affinity     : %s\n", (ENABLE_AFFINITY)	? "ON" : "OFF");
	printf(" Relaxed FIFOs: %s\n", (RELAX_FIFO_MAX_LEN) ? "ON" : "OFF");
	printf(" DGLs         : %s\n", (USE_DYNAMIC_GROUP_LOCKS) ? "ON" : "OFF");
	printf(" Chunking     : %s\n", (ENABLE_CHUNKING)	? "ON" : "OFF");
	printf(" Chunk Size   : %d\n", (int)CHUNK_SIZE);
	printf(" Engine Locks : %s\n", (NUM_SIMULT_USERS == 2) ? "ON" : "OFF");
	printf(" Mig Via SysM : %s\n", (MIGRATE_VIA_SYSMEM) ? "ON" : "OFF");
	printf(" Unconstrainted: %s\n", (UNCONSTRAINED_ACCESS) ? "ON" : "OFF");

	printf("\n");
#endif

	if (argc == 1) {
		cout << "No command arguments are provided." << endl;
		return false;
	}

	return true;	
}

bool read_config_file(const char* fname)
{	
	ifstream fin(fname);
	if (fin.fail()) {
		cout << "Error, unable to open " << fname << endl;
		return false;
	}
	
	string line, str;
	size_t p, q;
	while (!fin.eof()) {
		getline(fin, line);

		if(fin.eof())
			break;

		p = line.find_first_of('"'); 
		q = line.find_first_of('"', p+1);
		if (line[0] == '%' || line[0] == '#' || line[1] == '/') continue;

		istringstream ss(line);
		ss >> str;
		//cout << str << "|" << endl ;

		if (str == "GPU") {
			ss >> str;  b_gpu = (str.find("ON") != string::npos);
		}
		else if (str == "IMU") {
			ss >> str;  b_imu = (str.find("ON") != string::npos);
		}
		else if (str == "Log") {
			ss >> str;  b_log = (str.find("ON") != string::npos);
		}
		else if (str == "Stop") {
			ss >> str;  b_stop = (str.find("ON") != string::npos);
		}
		else if (str == "Display") {
			ss >> str;  b_display = (str.find("ON") != string::npos);
		}
		else if (str == "Time_offset") {
			ss >> str;  t_offset = atof(str.c_str());
		}
		else if (str == "Disp_scale") {
			ss >> str;  disp_scale = atof(str.c_str());
		}
		else if (str == "Feature_num") {
			ss >> str;  min_feature = max_feature = atoi(str.c_str());
		}
		else if (str == "Templ_size") {
			ss >> str;  templ_size = atoi(str.c_str());
		}
		else if (str == "Mask_size") {
			ss >> str;  mask_size = atoi(str.c_str());
		}
		else if (str == "Pyramid_level") {
			ss >> str;  pyramid_level = atoi(str.c_str());
		}
		else if (str == "Image_width") {
			ss >> str;  img_width = atoi(str.c_str());
		}
		else if (str == "Image_height") {
			ss >> str;  img_height = atoi(str.c_str());
		}
		else if (str == "Image_data") {
			if (p == string::npos || q == string::npos) {
				cout << "Error in setting Image_data in configuration file" << endl;
				return false;
			}
			line.copy(img_list_name, q-p-1, p+1);
		}
		else if (str == "IMU_data") {
			if (p == string::npos || q == string::npos) {
				cout << "Error in setting IMU_data in configuration file" << endl;
				return false;
			}
			line.copy(imu_list_name, q-p-1, p+1);
		}
		else if (str == "IMU_shape_param") {
			if (p == string::npos || q == string::npos) {
				cout << "Error in setting IMU_shape_param in configuration file" << endl;
				return false;
			}
			line.copy(imu_shape_name, q-p-1, p+1);
		}
		else if (str == "Camera_param") {
			if (p == string::npos || q == string::npos) {
				cout << "Error in setting Camera_param in configuration file" << endl;
				return false;
			}
			line.copy(camera_param_name, q-p-1, p+1);
		}
		else if (str == "Camera_table") {
			if (p == string::npos || q == string::npos) {
				cout << "Error in setting Camera_table in configuration file" << endl;
				return false;
			}
			line.copy(camera_table_name, q-p-1, p+1);
		}
		else if (str == "Log_filename") {
			if (p == string::npos || q == string::npos) {
				cout << "Error in setting Log_filename in configuration file" << endl;
				return false;
			}
			line.copy(track_log_name, q-p-1, p+1);
		}

		else if (str == "WCET_MS") {
			ss >> str;  wcet_ms = atof(str.c_str());
		}
		else if (str == "DesiredFrameRateHz") {
			ss >> str; period_ms = (1.0f/atof(str.c_str())) * 1000.f;
		}
		else if (str == "Num_GPUS") {
			ss >> str; numGPUs = atoi(str.c_str());
			GPU_PARTITION_SIZE = numGPUs;						
		}
		else if (str == "Num_Users") {
			ss >> str; numGPUUsers = atoi(str.c_str());
		}
		else if (str == "Wait_Release") {
			ss >> str;  b_waitForRelease = (str.find("ON") != string::npos);
		}
		else if (str == "Affinity") {
			ss >> str;  ENABLE_AFFINITY = (str.find("ON") != string::npos);
		}
		else if (str == "Relax_FIFO") {
			ss >> str;  RELAX_FIFO_MAX_LEN = (str.find("ON") != string::npos);
		}
		else if (str == "DGL") {
			ss >> str;  USE_DYNAMIC_GROUP_LOCKS = (str.find("ON") != string::npos);
		}
		else if (str == "Chunk_Enable") {
			ss >> str;  ENABLE_CHUNKING = (str.find("ON") != string::npos);
		}		
		else if (str == "Chunk_Size") {
			ss >> str;  CHUNK_SIZE = atof(str.c_str());
		}
		else if (str == "Sys_Mem") {
			ss >> str; MIGRATE_VIA_SYSMEM = (str.find("ON") != string::npos);
		}
		else if (str == "Engine_Locks") {
			ss >> str;  NUM_SIMULT_USERS = (str.find("ON") != string::npos) ? 2 : 1;
		}		
		else if (str == "Unconstrained_Access") {
			ss >> str;  UNCONSTRAINED_ACCESS = (str.find("ON") != string::npos);
		}
	}

	return true;
}
